# Generated by Django 3.1.1 on 2020-09-29 15:34

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='BlacklistedToken',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('username', models.CharField(db_index=True, editable=False, max_length=255)),
                ('token', models.TextField(editable=False)),
                ('expires_at', models.DateTimeField(db_index=True)),
                ('blacklisted_at', models.DateTimeField(db_index=True, editable=False)),
            ],
        ),
        # will only be applied if possible
        # migrations.AddConstraint(
        #     model_name='blacklistedtoken',
        #     constraint=models.UniqueConstraint(fields=('username', 'token'), name='unique_username_token'),
        # ),
    ]
